"""
The eoworkflow module, together with eotask and eodata, provides core building blocks for specifying and executing
workflows.

A workflow is a directed (acyclic) graph composed of instances of EOTask objects. Each task may take as input the
results of other tasks and external arguments. The external arguments are passed anew each time the workflow is
executed. The workflow builds the computational graph, performs dependency resolution, and executes the tasks.
If the input graph is cyclic, the workflow raises a `CyclicDependencyError`.

The result of a workflow execution is an immutable mapping from tasks to results. The result contains tasks with
zero out-degree (i.e. terminal tasks).

The workflow can be exported to a DOT description language and visualized.

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import collections
import logging
import warnings
import uuid
import copy

import attr

from .eotask import EOTask
from .graph import DirectedGraph


LOGGER = logging.getLogger(__name__)


class CyclicDependencyError(ValueError):
    """ This error is raised when trying to initialize `EOWorkflow` with a cyclic dependency graph
    """


class EOWorkflow:
    """ A basic eo-learn object for building workflows from a list of task dependencies

    Example:

        .. code-block:: python

            workflow = EOWorkflow([  # task1, task2, task3 are initialized EOTasks
                (task1, [], 'My first task'),
                (task2, []),
                (task3, [task1, task2], 'Task that depends on previous 2 tasks')
            ])
    """
    def __init__(self, dependencies, task_names=None):
        """
        :param dependencies: A list of dependencies between tasks, specifying the computational graph.
        :type dependencies: list(tuple or Dependency)
        """
        self.id_gen = _UniqueIdGenerator()

        if task_names:
            warnings.warn("Parameter 'task_names' will soon be removed. Everything can be specified with "
                          "'dependencies' parameter, including task names", DeprecationWarning, stacklevel=2)

        self.dependencies = self._parse_dependencies(dependencies, task_names)
        self.uuid_dict = self._set_task_uuid(self.dependencies)
        self.dag = self.create_dag(self.dependencies)
        self.ordered_dependencies = self._schedule_dependencies(self.dag)

    @staticmethod
    def _parse_dependencies(dependencies, task_names):
        """ Parses dependencies and adds names of task_names

        :param dependencies: Input of dependency parameter
        :type dependencies: list(tuple or Dependency)
        :param task_names: Human-readable names of tasks
        :type task_names: dict(EOTask: str) or None
        :return: List of dependencies
        :rtype: list(Dependency)
        """
        parsed_dependencies = [dep if isinstance(dep, Dependency) else Dependency(*dep) for dep in dependencies]
        for dep in parsed_dependencies:
            if task_names and dep.task in task_names:
                dep.set_name(task_names[dep.task])
        return parsed_dependencies

    def _set_task_uuid(self, dependencies):
        """ Adds universally unique user ids (UUID) to each task of the workflow

        :param dependencies: The list of dependencies between tasks defining the computational graph
        :type dependencies: list(Dependency)
        :return: A dictionary mapping UUID to dependencies
        :rtype: dict(str: Dependency)
        """
        uuid_dict = {}
        for dep in dependencies:
            task = dep.task
            if task.private_task_config.uuid in uuid_dict:
                raise ValueError('EOWorkflow cannot execute the same instance of EOTask multiple times')

            task.private_task_config.uuid = self.id_gen.get_next()
            uuid_dict[task.private_task_config.uuid] = dep

        return uuid_dict

    def create_dag(self, dependencies):
        """ Creates a directed graph from dependencies

        :param dependencies: A list of Dependency objects
        :type dependencies: list(Dependency)
        :return: A directed graph of the workflow
        :rtype: DirectedGraph
        """
        dag = DirectedGraph()
        for dep in dependencies:
            for vertex in dep.inputs:
                task_uuid = vertex.private_task_config.uuid
                if task_uuid not in self.uuid_dict:
                    raise ValueError('Task {}, which is an input of a task {}, is not part of the defined '
                                     'workflow'.format(vertex.__class__.__name__, dep.name))
                dag.add_edge(self.uuid_dict[task_uuid], dep)
            if not dep.inputs:
                dag.add_vertex(dep)
        return dag

    @staticmethod
    def _schedule_dependencies(dag):
        """ Computes an ordering < of tasks so that for any two tasks t and t' we have that if t depends on t' then
        t' < t. In words, all dependencies of a task precede the task in this ordering.

        :param dag: A directed acyclic graph representing dependencies between tasks.
        :type dag: DirectedGraph
        :return: A list of topologically ordered dependecies
        :rtype: list(Dependency)
        """
        in_degrees = dict(dag.get_indegrees())

        independent_vertices = collections.deque([vertex for vertex in dag if dag.get_indegree(vertex) == 0])
        topological_order = []
        while independent_vertices:
            v_vertex = independent_vertices.popleft()
            topological_order.append(v_vertex)

            for u_vertex in dag[v_vertex]:
                in_degrees[u_vertex] -= 1
                if in_degrees[u_vertex] == 0:
                    independent_vertices.append(u_vertex)

        if len(topological_order) != len(dag):
            raise CyclicDependencyError('Tasks do not form an acyclic graph')

        return topological_order

    def execute(self, input_args=None, monitor=False):
        """ Executes the workflow

        :param input_args: External input arguments to the workflow. They have to be in a form of a dictionary where
            each key is an EOTask used in the workflow and each value is a dictionary or a tuple of arguments.
        :type input_args: dict(EOTask: dict(str: object) or tuple(object))
        :param monitor: If True workflow execution will be monitored
        :type monitor: bool
        :return: An immutable mapping containing results of terminal tasks
        :rtype: WorkflowResults
        """
        out_degs = dict(self.dag.get_outdegrees())

        input_args = self.parse_input_args(input_args)

        results = WorkflowResults(self._execute_tasks(input_args=input_args, out_degs=out_degs, monitor=monitor))

        LOGGER.debug('Workflow finished with %s', repr(results))
        return results

    @staticmethod
    def parse_input_args(input_args):
        """ Parses EOWorkflow input arguments provided by user and raises an error if something is wrong. This is
        done automatically in the process of workflow execution
        """
        input_args = input_args if input_args else {}
        for task, args in input_args.items():
            if not isinstance(task, EOTask):
                raise ValueError('Invalid input argument {}, should be an instance of EOTask'.format(task))

            if not isinstance(args, (tuple, dict)):
                raise ValueError('Execution input arguments of each task should be a dictionary or a tuple, for task '
                                 '{} got arguments of type {}'.format(task.__class__.__name__, type(args)))

        return input_args

    def _execute_tasks(self, *, input_args, out_degs, monitor):
        """ Executes tasks comprising the workflow in the predetermined order

        :param input_args: External input arguments to the workflow.
        :type input_args: Dict
        :param out_degs: Dictionary mapping vertices (task IDs) to their out-degrees. (The out-degree equals the number
        of tasks that depend on this task.)
        :type out_degs: Dict
        :return: A dictionary mapping dependencies to task results
        :rtype: dict
        """
        intermediate_results = {}

        for dep in self.ordered_dependencies:
            result = self._execute_task(dependency=dep,
                                        input_args=input_args,
                                        intermediate_results=intermediate_results,
                                        monitor=monitor)

            intermediate_results[dep] = result

            self._relax_dependencies(dependency=dep,
                                     out_degrees=out_degs,
                                     intermediate_results=intermediate_results)

        return intermediate_results

    def _execute_task(self, *, dependency, input_args, intermediate_results, monitor):
        """ Executes a task of the workflow

        :param dependency: A workflow dependency
        :type dependency: Dependency
        :param input_args: External task parameters.
        :type input_args: dict
        :param intermediate_results: The dictionary containing intermediate results, including the results of all
        tasks that the current task depends on.
        :type intermediate_results: dict
        :return: The result of the task in dependency
        :rtype: object
        """
        task = dependency.task
        inputs = tuple(intermediate_results[self.uuid_dict[input_task.private_task_config.uuid]]
                       for input_task in dependency.inputs)

        kw_inputs = input_args.get(task, {})
        if isinstance(kw_inputs, tuple):
            inputs += kw_inputs
            kw_inputs = {}

        LOGGER.debug("Computing %s(*%s, **%s)", task.__class__.__name__, str(inputs), str(kw_inputs))
        return task(*inputs, **kw_inputs, monitor=monitor)

    def _relax_dependencies(self, *, dependency, out_degrees, intermediate_results):
        """ Relaxes dependencies incurred by ``task_id``. After the task with ID ``task_id`` has been successfully
        executed, all the tasks it depended on are updated. If ``task_id`` was the last remaining dependency of a task
        ``t`` then ``t``'s result is removed from memory and, depending on ``remove_intermediate``, from disk.

        :param dependency: A workflow dependency
        :type dependency: Dependency
        :param out_degrees: Out-degrees of tasks
        :type out_degrees: dict
        :param intermediate_results: The dictionary containing the intermediate results (needed by tasks that have yet
        to be executed) of the already-executed tasks
        :type intermediate_results: dict
        """
        current_task = dependency.task
        for input_task in dependency.inputs:
            dep = self.uuid_dict[input_task.private_task_config.uuid]
            out_degrees[dep] -= 1

            if out_degrees[dep] == 0:
                LOGGER.debug("Removing intermediate result for %s", current_task.__class__.__name__)
                del intermediate_results[dep]

    def get_tasks(self):
        """ Returns an ordered dictionary {task_name: task} of all tasks within this workflow

        :return: Ordered dictionary with key being task_name (str) and an instance of a corresponding task from this
            workflow. The order of tasks is the same as in which they will be executed.
        :rtype: OrderedDict
        """
        task_dict = collections.OrderedDict()
        for dep in self.ordered_dependencies:
            task_name = dep.name

            if task_name in task_dict:
                count = 0
                while dep.get_custom_name(count) in task_dict:
                    count += 1

                task_name = dep.get_custom_name(count)

            task_dict[task_name] = dep.task

        return task_dict

    def get_dot(self):
        """ Generates the DOT description of the underlying computational graph

        :return: The DOT representation of the computational graph
        :rtype: Digraph
        """
        visualization = self._get_visualization()
        return visualization.get_dot()

    def dependency_graph(self, filename=None):
        """ Visualize the computational graph

        :param filename: Filename of the output image together with file extension. Supported formats: `png`, `jpg`,
            `pdf`, ... . Check `graphviz` Python package for more options
        :type filename: str
        :return: The DOT representation of the computational graph, with some more formatting
        :rtype: Digraph
        """
        visualization = self._get_visualization()
        return visualization.dependency_graph(filename=filename)

    def _get_visualization(self):
        """ Helper method which provides EOWorkflowVisualization object
        """
        # pylint: disable=import-outside-toplevel,raise-missing-from
        try:
            from eolearn.visualization import EOWorkflowVisualization
        except ImportError:
            raise RuntimeError('Subpackage eo-learn-visualization has to be installed in order to use EOWorkflow '
                               'visualization methods')
        return EOWorkflowVisualization(self)


class LinearWorkflow(EOWorkflow):
    """ A linear version of EOWorkflow where each tasks only gets results of the previous task

    Example:

        .. code-block:: python

            workflow = LinearWorkflow(task1, task2, task3)
    """
    def __init__(self, *tasks, **kwargs):
        """
        :param tasks: Tasks in the order of execution
        :type tasks: EOTask
        """
        tasks = [self._parse_task(task) for task in tasks]
        tasks = self._make_tasks_unique(tasks)

        dependencies = [(task, [tasks[idx - 1][0]] if idx > 0 else [], name) for idx, (task, name) in enumerate(tasks)]
        super().__init__(dependencies, **kwargs)

    @staticmethod
    def _parse_task(task):
        """ Parses input task
        """
        if isinstance(task, EOTask):
            return task, None
        if isinstance(task, (tuple, list)) and len(task) == 2:
            return task

        raise ValueError('Cannot parse {}, expected an instance of EOTask or a tuple (EOTask, name)'.format(task))

    @staticmethod
    def _make_tasks_unique(tasks):
        """ If the same instances of tasks are given to LinearWorkflow, this will deep copy them
        """
        unique_tasks = []
        prev_tasks = set()

        for task, name in tasks:
            if task in prev_tasks:
                task = copy.deepcopy(task)

            prev_tasks.add(task)
            unique_tasks.append((task, name))

        return unique_tasks


@attr.s(eq=False)  # eq=False preserves the original hash
class Dependency:
    """ Class representing a node in EOWorkflow graph

    :param task: An instance of EOTask
    :type task: EOTask
    :param inputs: A list of EOTask instances which are dependencies of the given `task`
    :type inputs: list(EOTask) or EOTask
    :param name: Name of the Dependency node
    :type name: str or None
    """
    task = attr.ib(default=None)  # validator parameter could be used, but its error msg is ugly
    inputs = attr.ib(factory=list)
    name = attr.ib(default=None)

    def __attrs_post_init__(self):
        """ This is executed right after init method
        """
        if not isinstance(self.task, EOTask):
            raise ValueError('Value {} should be an instance of {}'.format(self.task, EOTask.__name__))
        self.task = self.task

        if isinstance(self.inputs, EOTask):
            self.inputs = [self.inputs]
        if not isinstance(self.inputs, (list, tuple)):
            raise ValueError('Value {} should be a list'.format(self.inputs))
        for input_task in self.inputs:
            if not isinstance(input_task, EOTask):
                raise ValueError('Value {} should be an instance of {}'.format(input_task, EOTask.__name__))

        if self.name is None:
            self.name = self.task.__class__.__name__

    def set_name(self, name):
        """ Sets a new name
        """
        self.name = name

    def get_custom_name(self, number=0):
        """ Provides custom task name according to given number. E.g. FooTask -> FooTask
        """
        if number:
            return '{}_{}'.format(self.name, number)
        return self.name


class WorkflowResults(collections.abc.Mapping):
    """ The result of a workflow is an (immutable) dictionary mapping [1] from EOTasks to results of the workflow.

    When an EOTask is passed as an index, its UUID, assigned during workflow execution, is used as the key (as opposed
    to the result of invoking __repr__ on the EO task). This ensures that indexing by task works even after pickling,
    and makes dealing with checkpoints more convenient.

    [1] https://docs.python.org/3.6/library/collections.abc.html#collections-abstract-base-classes
    """
    def __init__(self, results):
        self._result = results
        self._uuid_dict = {dep.task.private_task_config.uuid: dep for dep in results}

    def __getitem__(self, item):
        if isinstance(item, EOTask):
            item = self._uuid_dict[item.private_task_config.uuid]
        return self._result[item]

    def __len__(self):
        return len(self._result)

    def __iter__(self):
        return iter(self._result)

    def __contains__(self, item):
        if isinstance(item, EOTask):
            item = self._uuid_dict.get(item.private_task_config.uuid, item)
        return item in self._result

    def __eq__(self, other):
        return self._result == other

    def __ne__(self, other):
        return self._result != other

    def keys(self):
        """ Returns dictionary keys """
        return {dep.task: None for dep in self._result}.keys()

    def values(self):
        """ Returns dictionary values """
        return self._result.values()

    def items(self):
        """ Returns dictionary items """
        return {dep.task: value for dep, value in self._result.items()}.items()

    def get(self, key, default=None):
        """ Dictionary get method """
        if isinstance(key, EOTask):
            key = self._uuid_dict[key.private_task_config.uuid]
        return self._result.get(key, default)

    def eopatch(self):
        """ Return the EOPatch from the workflow result """
        return list(self.values())[-1]

    def __repr__(self):
        repr_list = ['{}('.format(self.__class__.__name__)]

        for _, dep in self._uuid_dict.items():
            result_repr = repr(self._result[dep]).replace('\n', '\n    ')
            dependency_repr = '{}({}):\n    {}'.format(Dependency.__name__, dep.name, result_repr)

            repr_list.append(dependency_repr)

        return '\n  '.join(repr_list) + '\n)'


class _UniqueIdGenerator:
    """ Generator of unique IDs, which is used in workflows only
    """

    MAX_UUIDS = 2 ** 20

    def __init__(self):
        self.uuids = set()

    def _next(self):
        if len(self.uuids) + 1 > _UniqueIdGenerator.MAX_UUIDS:
            raise MemoryError('Limit of max UUIDs reached')

        while True:
            uid = uuid.uuid4()
            if uid not in self.uuids:
                self.uuids.add(uid)
                return uid

    def get_next(self):
        """ Generates an ID
        """
        return self._next().hex



"""
The module handles execution and monitoring of workflows. It enables executing a workflow multiple times and in
parallel. It monitors execution times and handles any error that might occur in the process. At the end it generates a
report which contains summary of the workflow and process of execution.

All this is implemented in EOExecutor class.

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import os
import logging
import threading
import traceback
import concurrent.futures
import datetime as dt
import multiprocessing
import warnings

from tqdm.auto import tqdm

from .eoworkflow import EOWorkflow

from .utilities import LogFileFilter

LOGGER = logging.getLogger(__name__)

try:
    MULTIPROCESSING_LOCK = multiprocessing.Manager().Lock()
except BaseException:
    MULTIPROCESSING_LOCK = None


class EOExecutor:
    """ Simultaneously executes a workflow with different input arguments. In the process it monitors execution and
    handles errors. It can also save logs and create a html report about each execution.
    """
    REPORT_FILENAME = 'report.html'

    STATS_START_TIME = 'start_time'
    STATS_END_TIME = 'end_time'
    STATS_ERROR = 'error'
    RESULTS = 'results'

    def __init__(self, workflow, execution_args, *, save_logs=False, logs_folder='.', logs_filter=None,
                 execution_names=None):
        """
        :param workflow: A prepared instance of EOWorkflow class
        :type workflow: EOWorkflow
        :param execution_args: A list of dictionaries where each dictionary represents execution inputs for the
            workflow. `EOExecutor` will execute the workflow for each of the given dictionaries in the list. The
            content of such dictionary will be used as `input_args` parameter in `EOWorkflow.execution` method.
            Check `EOWorkflow.execution` for definition of a dictionary structure.
        :type execution_args: list(dict(EOTask: dict(str: object) or tuple(object)))
        :param save_logs: Flag used to specify if execution log files should be saved locally on disk
        :type save_logs: bool
        :param logs_folder: A folder where logs and execution report should be saved
        :type logs_folder: str
        :param logs_filter: An instance of a custom filter object that will filter certain logs from being written into
            logs. It works only if save_logs parameter is set to True.
        :type logs_filter: logging.Filter or None
        :param execution_names: A list of execution names, which will be shown in execution report
        :type execution_names: list(str) or None
        """
        self.workflow = workflow
        self.execution_args = self._parse_execution_args(execution_args)
        self.save_logs = save_logs
        self.logs_folder = logs_folder
        self.logs_filter = logs_filter
        self.execution_names = self._parse_execution_names(execution_names, self.execution_args)

        self.start_time = None
        self.report_folder = None
        self.general_stats = {}
        self.execution_logs = None
        self.execution_stats = None

    @staticmethod
    def _parse_execution_args(execution_args):
        """ Parses execution arguments provided by user and raises an error if something is wrong
        """
        if not isinstance(execution_args, (list, tuple)):
            raise ValueError("Parameter 'execution_args' should be a list")

        return [EOWorkflow.parse_input_args(input_args) for input_args in execution_args]

    @staticmethod
    def _parse_execution_names(execution_names, execution_args):
        """ Parses a list of execution names
        """
        if execution_names is None:
            return [str(num) for num in range(1, len(execution_args) + 1)]

        if not isinstance(execution_names, (list, tuple)) or len(execution_names) != len(execution_args):
            raise ValueError("Parameter 'execution_names' has to be a list of the same size as the list of "
                             "execution arguments")
        return execution_names

    def run(self, workers=1, multiprocess=True, return_results=False):
        """ Runs the executor with n workers.

        :param workers: Maximum number of workflows which will be executed in parallel. Default value is `1` which will
            execute workflows consecutively. If set to `None` the number of workers will be the number of processors
            of the system.
        :type workers: int or None
        :param multiprocess: If `True` it will use `concurrent.futures.ProcessPoolExecutor` which will distribute
            workflow executions among multiple processors. If `False` it will use
            `concurrent.futures.ThreadPoolExecutor` which will distribute workflow among multiple threads.
            However even when `multiprocess=False`, tasks from workflow could still be using multiple processors.
            This parameter is used especially because certain task cannot run with
            `concurrent.futures.ProcessPoolExecutor`.
            In case of `workers=1` this parameter is ignored and workflows will be executed consecutively.
        :type multiprocess: bool
        :param return_results: If `True` this method will return a list of all results of the execution. Note that
            this might exceed the available memory. By default this parameter is set to `False`.
        :type: bool
        :return: If `return_results` is set to `True` it will return a list of results, otherwise it will return `None`
        :rtype: None or list(eolearn.core.WorkflowResults)
        """
        self.start_time = dt.datetime.now()
        self.report_folder = self._get_report_folder()
        if self.save_logs and not os.path.isdir(self.report_folder):
            os.mkdir(self.report_folder)

        execution_num = len(self.execution_args)
        log_paths = self._get_log_paths()

        filter_logs_by_thread = not multiprocess and workers > 1
        processing_args = [(self.workflow, init_args, log_path, return_results, filter_logs_by_thread)
                           for init_args, log_path in zip(self.execution_args, log_paths)]

        if workers == 1:
            processing_type = 'single process'
            self.execution_stats = list(tqdm(map(self._execute_workflow, processing_args), total=len(processing_args)))
        else:
            if multiprocess:
                pool_executor_class = concurrent.futures.ProcessPoolExecutor
                processing_type = 'multiprocessing'
            else:
                pool_executor_class = concurrent.futures.ThreadPoolExecutor
                processing_type = 'multithreading'

            with pool_executor_class(max_workers=workers) as executor:
                self.execution_stats = list(tqdm(executor.map(self._execute_workflow, processing_args),
                                                 total=len(processing_args)))

        self.general_stats = self._prepare_general_stats(workers, processing_type)

        self.execution_logs = [None] * execution_num
        if self.save_logs:
            for idx, log_path in enumerate(log_paths):
                with open(log_path) as fin:
                    self.execution_logs[idx] = fin.read()

        return [stats.get(self.RESULTS) for stats in self.execution_stats] if return_results else None

    def _try_add_logging(self, log_path, filter_logs_by_thread):
        """ Adds a handler to a logger and returns them both. In case this fails it shows a warning.
        """
        if log_path:
            try:
                logger = logging.getLogger()
                logger.setLevel(logging.DEBUG)
                handler = self._get_log_handler(log_path, filter_logs_by_thread)
                logger.addHandler(handler)
                return logger, handler
            except BaseException as exception:
                warnings.warn('Failed to create logs with exception: {}'.format(repr(exception)),
                              category=RuntimeWarning)

        return None, None

    @classmethod
    def _try_remove_logging(cls, log_path, logger, handler, stats):
        """ Removes a handler from a logger in case that handler exists.
        """
        if log_path:
            try:
                message = 'EOWorkflow execution {}'.format('failed' if cls.STATS_ERROR in stats else 'finished')
                logger.debug(message)
                handler.close()
                logger.removeHandler(handler)
            except BaseException:
                pass

    def _execute_workflow(self, process_args):
        """ Handles a single execution of a workflow
        """
        workflow, input_args, log_path, return_results, filter_logs_by_thread = process_args
        logger, handler = self._try_add_logging(log_path, filter_logs_by_thread)
        stats = {self.STATS_START_TIME: dt.datetime.now()}
        try:
            results = workflow.execute(input_args, monitor=True)

            if return_results:
                stats[self.RESULTS] = results

        except KeyboardInterrupt as exception:
            raise KeyboardInterrupt from exception
        except BaseException:
            stats[self.STATS_ERROR] = traceback.format_exc()
        stats[self.STATS_END_TIME] = dt.datetime.now()

        self._try_remove_logging(log_path, logger, handler, stats)

        return stats

    def _get_log_handler(self, log_path, filter_logs_by_thread):
        """ Provides object which handles logs
        """
        handler = logging.FileHandler(log_path)
        formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        handler.setFormatter(formatter)

        if filter_logs_by_thread:
            handler.addFilter(LogFileFilter(threading.currentThread().getName()))

        if self.logs_filter:
            handler.addFilter(self.logs_filter)

        return handler

    def _prepare_general_stats(self, workers, processing_type):
        """ Prepares a dictionary with a general statistics about executions
        """
        failed_count = sum(self.STATS_ERROR in stats for stats in self.execution_stats)
        return {
            self.STATS_START_TIME: self.start_time,
            self.STATS_END_TIME: dt.datetime.now(),
            'finished': len(self.execution_stats) - failed_count,
            'failed': failed_count,
            'processing_type': processing_type,
            'workers': workers
        }

    def _get_report_folder(self):
        """ Returns file path of folder where report will be saved
        """
        return os.path.join(self.logs_folder,
                            'eoexecution-report-{}'.format(self.start_time.strftime("%Y_%m_%d-%H_%M_%S")))

    def _get_log_paths(self):
        """ Returns a list of file paths containing logs
        """
        if self.save_logs:
            return [os.path.join(self.report_folder, 'eoexecution-{}.log'.format(name))
                    for name in self.execution_names]

        return [None] * len(self.execution_names)

    def get_successful_executions(self):
        """ Returns a list of IDs of successful executions. The IDs are integers from interval
        `[0, len(execution_args) - 1]`, sorted in increasing order.

        :return: List of successful execution IDs
        :rtype: list(int)
        """
        return [idx for idx, stats in enumerate(self.execution_stats) if self.STATS_ERROR not in stats]

    def get_failed_executions(self):
        """ Returns a list of IDs of failed executions. The IDs are integers from interval
        `[0, len(execution_args) - 1]`, sorted in increasing order.

        :return: List of failed execution IDs
        :rtype: list(int)
        """
        return [idx for idx, stats in enumerate(self.execution_stats) if self.STATS_ERROR in stats]

    def get_report_filename(self):
        """ Returns the filename and file path of the report

        :return: Report filename
        :rtype: str
        """
        return os.path.join(self.report_folder, self.REPORT_FILENAME)

    def make_report(self):
        """ Makes a html report and saves it into the same folder where logs are stored.
        """
        # pylint: disable=import-outside-toplevel,raise-missing-from
        try:
            from eolearn.visualization import EOExecutorVisualization
        except ImportError:
            raise RuntimeError('Subpackage eo-learn-visualization has to be installed in order to create EOExecutor '
                               'reports')

        return EOExecutorVisualization(self).make_report()


def execute_with_mp_lock(execution_function, *args, **kwargs):
    """ A helper utility function that executes a given function with multiprocessing lock if the process is being
    executed in a multi-processing mode

    :param execution_function: A function
    :param args: Function's positional arguments
    :param kwargs: Function's keyword arguments
    :return: Function's results
    """
    if multiprocessing.current_process().name == 'MainProcess' or MULTIPROCESSING_LOCK is None:
        return execution_function(*args, **kwargs)

    with MULTIPROCESSING_LOCK:
        return execution_function(*args, **kwargs)



"""
A module implementing EOPatch merging utility

Credits:
Copyright (c) 2018-2020 William Ouellette
Copyright (c) 2017-2020 Matej Aleksandrov, Matej Batič, Grega Milčinski, Matic Lubej, Devis Peresutti (Sinergise)
Copyright (c) 2017-2020 Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import functools
import warnings
from collections.abc import Callable

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame

from .constants import FeatureType
from .utilities import FeatureParser


def merge_eopatches(*eopatches, features=..., time_dependent_op=None, timeless_op=None):
    """ Merge features of given EOPatches into a new EOPatch

    :param eopatches: Any number of EOPatches to be merged together
    :type eopatches: EOPatch
    :param features: A collection of features to be merged together. By default all features will be merged.
    :type features: object
    :param time_dependent_op: An operation to be used to join data for any time-dependent raster feature. Before
        joining time slices of all arrays will be sorted. Supported options are:
        - None (default): If time slices with matching timestamps have the same values, take one. Raise an error
          otherwise.
        - 'concatenate': Keep all time slices, even the ones with matching timestamps
        - 'min': Join time slices with matching timestamps by taking minimum values. Ignore NaN values.
        - 'max': Join time slices with matching timestamps by taking maximum values. Ignore NaN values.
        - 'mean': Join time slices with matching timestamps by taking mean values. Ignore NaN values.
        - 'median': Join time slices with matching timestamps by taking median values. Ignore NaN values.
    :type time_dependent_op: str or Callable or None
    :param timeless_op: An operation to be used to join data for any timeless raster feature. Supported options
        are:
        - None (default): If arrays are the same, take one. Raise an error otherwise.
        - 'concatenate': Join arrays over the last (i.e. bands) dimension
        - 'min': Join arrays by taking minimum values. Ignore NaN values.
        - 'max': Join arrays by taking maximum values. Ignore NaN values.
        - 'mean': Join arrays by taking mean values. Ignore NaN values.
        - 'median': Join arrays by taking median values. Ignore NaN values.
    :type timeless_op: str or Callable or None
    :return: A dictionary with EOPatch features and values
    :rtype: Dict[(FeatureType, str), object]
    """
    reduce_timestamps = time_dependent_op != 'concatenate'
    time_dependent_op = _parse_operation(time_dependent_op, is_timeless=False)
    timeless_op = _parse_operation(timeless_op, is_timeless=True)

    all_features = {feature for eopatch in eopatches for feature in FeatureParser(features)(eopatch)}
    eopatch_content = {}

    timestamps, sort_mask, split_mask = _merge_timestamps(eopatches, reduce_timestamps)
    eopatch_content[FeatureType.TIMESTAMP] = timestamps

    for feature in all_features:
        feature_type, feature_name = feature

        if feature_type.is_raster():
            if feature_type.is_time_dependent():
                eopatch_content[feature] = _merge_time_dependent_raster_feature(
                    eopatches, feature, time_dependent_op, sort_mask, split_mask
                )
            else:
                eopatch_content[feature] = _merge_timeless_raster_feature(eopatches, feature,
                                                                          timeless_op)

        if feature_type.is_vector():
            eopatch_content[feature] = _merge_vector_feature(eopatches, feature)

        if feature_type is FeatureType.META_INFO:
            eopatch_content[feature] = _select_meta_info_feature(eopatches, feature_name)

        if feature_type is FeatureType.BBOX:
            eopatch_content[feature] = _get_common_bbox(eopatches)

    return eopatch_content


def _parse_operation(operation_input, is_timeless):
    """ Transforms operation's instruction (i.e. an input string) into a function that can be applied to a list of
    arrays. If the input already is a function it returns it.
    """
    if isinstance(operation_input, Callable):
        return operation_input

    try:
        return {
            None: _return_if_equal_operation,
            'concatenate': functools.partial(np.concatenate, axis=-1 if is_timeless else 0),
            'mean': functools.partial(np.nanmean, axis=0),
            'median': functools.partial(np.nanmedian, axis=0),
            'min': functools.partial(np.nanmin, axis=0),
            'max': functools.partial(np.nanmax, axis=0)
        }[operation_input]
    except KeyError as exception:
        raise ValueError(f'Merge operation {operation_input} is not supported') from exception


def _return_if_equal_operation(arrays):
    """ Checks if arrays are all equal and returns first one of them. If they are not equal it raises an error.
    """
    if _all_equal(arrays):
        return arrays[0]
    raise ValueError('Cannot merge given arrays because their values are not the same')


def _merge_timestamps(eopatches, reduce_timestamps):
    """ Merges together timestamps from EOPatches. It also prepares masks on how to sort and join data in any
    time-dependent raster feature.
    """
    all_timestamps = [timestamp for eopatch in eopatches for timestamp in eopatch.timestamp
                      if eopatch.timestamp is not None]

    if not all_timestamps:
        return [], None, None

    sort_mask = np.argsort(all_timestamps)
    all_timestamps = sorted(all_timestamps)

    if not reduce_timestamps:
        return all_timestamps, sort_mask, None

    split_mask = [
        index + 1 for index, (timestamp, next_timestamp) in enumerate(zip(all_timestamps[:-1], all_timestamps[1:]))
        if timestamp != next_timestamp
    ]
    reduced_timestamps = [timestamp for index, timestamp in enumerate(all_timestamps)
                          if index == 0 or timestamp != all_timestamps[index - 1]]

    return reduced_timestamps, sort_mask, split_mask


def _merge_time_dependent_raster_feature(eopatches, feature, operation, sort_mask, split_mask):
    """ Merges numpy arrays of a time-dependent raster feature with a given operation and masks on how to sort and join
    time raster's time slices.
    """
    arrays = _extract_feature_values(eopatches, feature)

    merged_array = np.concatenate(arrays, axis=0)
    del arrays

    if sort_mask is None:
        return merged_array
    merged_array = merged_array[sort_mask]

    if split_mask is None or len(split_mask) == merged_array.shape[0] - 1:
        return merged_array

    split_arrays = np.split(merged_array, split_mask)
    del merged_array

    try:
        split_arrays = [operation(array_chunk) for array_chunk in split_arrays]
    except ValueError as exception:
        raise ValueError(f'Failed to merge {feature} with {operation}, try setting a different value for merging '
                         f'parameter time_dependent_op') from exception

    return np.array(split_arrays)


def _merge_timeless_raster_feature(eopatches, feature, operation):
    """ Merges numpy arrays of a timeless raster feature with a given operation.
    """
    arrays = _extract_feature_values(eopatches, feature)

    if len(arrays) == 1:
        return arrays[0]

    try:
        return operation(arrays)
    except ValueError as exception:
        raise ValueError(f'Failed to merge {feature} with {operation}, try setting a different value for merging '
                         f'parameter timeless_op') from exception


def _merge_vector_feature(eopatches, feature):
    """ Merges GeoDataFrames of a vector feature.
    """
    dataframes = _extract_feature_values(eopatches, feature)

    if len(dataframes) == 1:
        return dataframes[0]

    crs_list = [dataframe.crs for dataframe in dataframes if dataframe.crs is not None]
    if not crs_list:
        crs_list = [None]
    if not _all_equal(crs_list):
        raise ValueError(f'Cannot merge feature {feature} because dataframes are defined for '
                         f'different CRS')

    merged_dataframe = GeoDataFrame(pd.concat(dataframes, ignore_index=True), crs=crs_list[0])
    merged_dataframe = merged_dataframe.drop_duplicates(ignore_index=True)
    # In future a support for vector operations could be added here

    return merged_dataframe


def _select_meta_info_feature(eopatches, feature_name):
    """ Selects a value for a meta info feature of a merged EOPatch. By default the value is the first one.
    """
    values = _extract_feature_values(eopatches, (FeatureType.META_INFO, feature_name))

    if not _all_equal(values):
        message = f'EOPatches have different values of meta info feature {feature_name}. The first value will be ' \
                  f'used in a merged EOPatch'
        warnings.warn(message, category=UserWarning)

    return values[0]


def _get_common_bbox(eopatches):
    """ Makes sure that all EOPatches, which define a bounding box and CRS, define the same ones.
    """
    bboxes = [eopatch.bbox for eopatch in eopatches if eopatch.bbox is not None]

    if not bboxes:
        return None

    if _all_equal(bboxes):
        return bboxes[0]
    raise ValueError('Cannot merge EOPatches because they are defined for different bounding boxes')


def _extract_feature_values(eopatches, feature):
    """ A helper function that extracts a feature values from those EOPatches where a feature exists.
    """
    feature_type, feature_name = feature
    return [eopatch[feature] for eopatch in eopatches if feature_name in eopatch[feature_type]]


def _all_equal(values):
    """ A helper function that checks if all values in a given list are equal to each other.
    """
    first_value = values[0]

    if isinstance(first_value, np.ndarray):
        is_numeric_dtype = np.issubdtype(first_value.dtype, np.number)
        return all(np.array_equal(first_value, array, equal_nan=is_numeric_dtype) for array in values[1:])

    return all(first_value == value for value in values[1:])



"""
This module implements the core class hierarchy for implementing EO tasks. An EO task is any class the inherits
from the abstract EOTask class. Each EO task has to implement the execute method; invoking __call__ on a EO task
instance invokes the execute method. EO tasks are meant primarily to operate on EO patches (i.e. instances of EOPatch).

EO task classes are generally lightweight (i.e. not too complicated), short, and do one thing well. For example, an
EO task might take as input an EOPatch containing cloud mask and return as a result the cloud coverage for that mask.

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import sys
import logging
import datetime
import inspect
from collections import OrderedDict
from abc import ABC, abstractmethod

import attr

from .utilities import FeatureParser

LOGGER = logging.getLogger(__name__)


class EOTask(ABC):

    """Base class for EOTask."""
    def __new__(cls, *args, **kwargs):
        """Stores initialization parameters and the order to the instance attribute `init_args`."""
        self = super().__new__(cls)

        init_args = OrderedDict()
        for arg, value in zip(inspect.getfullargspec(self.__init__).args[1: len(args) + 1], args):
            init_args[arg] = repr(value)
        for arg in inspect.getfullargspec(self.__init__).args[len(args) + 1:]:
            if arg in kwargs:
                init_args[arg] = repr(kwargs[arg])

        self.private_task_config = _PrivateTaskConfig(init_args=init_args)

        return self

    def __mul__(self, other):
        """Creates a composite task of this and passed task."""
        return CompositeTask(other, self)

    def __call__(self, *eopatches, monitor=False, **kwargs):
        """Executes the task."""
        # if monitor:
        #     return self.execute_and_monitor(*eopatches, **kwargs)

        return self._execute_handling(*eopatches, **kwargs)

    def execute_and_monitor(self, *eopatches, **kwargs):
        """ In the current version nothing additional happens in this method
        """
        return self._execute_handling(*eopatches, **kwargs)

    def _execute_handling(self, *eopatches, **kwargs):
        """ Handles measuring execution time and error propagation
        """
        self.private_task_config.start_time = datetime.datetime.now()

        try:
            return_value = self.execute(*eopatches, **kwargs)
            self.private_task_config.end_time = datetime.datetime.now()
            return return_value
        except BaseException as exception:
            traceback = sys.exc_info()[2]

            # Some special exceptions don't accept an error message as a parameter and raise a TypeError in such case.
            try:
                errmsg = 'During execution of task {}: {}'.format(self.__class__.__name__, exception)
                extended_exception = type(exception)(errmsg)
            except TypeError:
                extended_exception = exception

            raise extended_exception.with_traceback(traceback)

    @abstractmethod
    def execute(self, *eopatches, **kwargs):
        """ Implement execute function
        """
        raise NotImplementedError

    @staticmethod
    def _parse_features(features, new_names=False, rename_function=None, default_feature_type=None,
                        allowed_feature_types=None):
        """ See eolearn.core.utilities.FeatureParser class.
        """
        return FeatureParser(features, new_names=new_names, rename_function=rename_function,
                             default_feature_type=default_feature_type, allowed_feature_types=allowed_feature_types)


@attr.s(eq=False)
class _PrivateTaskConfig:
    """ A container for general EOTask parameters required during EOWorkflow and EOExecution

    :param init_args: A dictionary of parameters and values used for EOTask initialization
    :type init_args: OrderedDict
    :param uuid: An unique hexadecimal identifier string a task gets in EOWorkflow
    :type uuid: str or None
    :param start_time: Time when task execution started
    :type start_time: datetime.datetime or None
    :param end_time: Time when task execution ended
    :type end_time: datetime.datetime or None
    """
    init_args = attr.ib()
    uuid = attr.ib(default=None)
    start_time = attr.ib(default=None)
    end_time = attr.ib(default=None)

    def __add__(self, other):
        return _PrivateTaskConfig(init_args=OrderedDict(list(self.init_args.items()) + list(other.init_args.items())))


class CompositeTask(EOTask):
    """Creates a task that is composite of two tasks.

    Note: Instead of directly using this task it might be more convenient to use `'*'` operation between tasks.
    Example: `composite_task = task1 * task2`

    :param eotask1: Task which will be executed first
    :type eotask1: EOTask
    :param eotask2: Task which will be executed on results of first task
    :type eotask2: EOTask
    """
    def __init__(self, eotask1, eotask2):
        self.eotask1 = eotask1
        self.eotask2 = eotask2

        self.private_task_config = eotask1.private_task_config + eotask2.private_task_config

    def execute(self, *eopatches, **kwargs):
        return self.eotask2.execute(self.eotask1.execute(*eopatches, **kwargs))



"""
A module implementing utilities for working with different filesystems

Credits:
Copyright (c) 2017-2020 Matej Aleksandrov, Matej Batič, Grega Milčinski, Matic Lubej, Devis Peresutti (Sinergise)
Copyright (c) 2017-2020 Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import os
from pathlib import Path, PurePath

import fs
from fs_s3fs import S3FS

from sentinelhub import SHConfig


def get_filesystem(path, create=False, config=None, **kwargs):
    """ A utility function for initializing any type of filesystem object with PyFilesystem2 package

    :param path: A filesystem path
    :type path: str
    :param create: If the filesystem path doesn't exist this flag indicates to either create it or raise an error
    :type create: bool
    :param config: A configuration object with AWS credentials
    :type config: SHConfig
    :param kwargs: Any keyword arguments to be passed forward
    :return: A filesystem object
    :rtype: fs.FS
    """
    if isinstance(path, Path):
        path = str(path)

    if path.startswith('s3://'):
        return load_s3_filesystem(path, config=config, **kwargs)

    return fs.open_fs(path, create=create, **kwargs)


def get_base_filesystem_and_path(*path_parts, **kwargs):
    """ Parses multiple strings that define a filesystem path and returns a filesystem object with a relative path
    on the filesystem

    :param path_parts: One or more strings defining a filesystem path
    :type path_parts: str
    :param kwargs: Parameters passed to get_filesystem function
    :return: A filesystem object and a relative path
    :rtype: (fs.FS, str)
    """
    path_parts = [str(part) for part in path_parts if part is not None]
    base_path = path_parts[0]

    if '://' in base_path:
        base_path_parts = base_path.split('/', 3)
        filesystem_path = '/'.join(base_path_parts[:-1])
        relative_path = '/'.join([base_path_parts[-1], *path_parts[1:]])

        return get_filesystem(filesystem_path, **kwargs), relative_path

    entire_path = os.path.abspath(os.path.join(*path_parts))
    pure_path = PurePath(entire_path)
    posix_path = pure_path.relative_to(pure_path.anchor).as_posix()

    return get_filesystem('/', **kwargs), posix_path


def load_s3_filesystem(path, strict=False, config=None):
    """ Loads AWS s3 filesystem from a path

    :param path: A path to a folder on s3 bucket that will be the base folder in this filesystem
    :type path: str
    :param strict: If `True` the filesystem will be making additional checks to the s3. Default is `False`.
    :type strict: bool
    :param config: A configuration object with AWS credentials. By default is set to None and in this case the default
        configuration will be taken.
    :type config: SHConfig or None
    :return: A S3 filesystem object
    :rtype: fs_s3fs.S3FS
    """
    if not path.startswith('s3://'):
        raise ValueError("AWS path has to start with s3:// but found '{}'".format(path))

    if config is None:
        config = SHConfig()

    path_chunks = path.split('/', 3)[2:]
    bucket_name = path_chunks[0]
    dir_path = path_chunks[1] if len(path_chunks) > 1 else '/'

    return S3FS(
        bucket_name=bucket_name,
        dir_path=dir_path,
        aws_access_key_id=config.aws_access_key_id if config.aws_access_key_id else None,
        aws_secret_access_key=config.aws_secret_access_key if config.aws_secret_access_key else None,
        strict=strict
    )
